/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    VisualizeUtils.java
 *    Copyright (C) 2000 Mark Hall
 *
 */

package weka.gui.visualize;

import weka.core.Utils;
import java.util.Properties;
import java.io.FileInputStream;

import java.awt.Color;
import javax.swing.JOptionPane;


/**
 * This class contains utility routines for visualization
 * 
 * @author Mark Hall (mhall@cs.waikato.ac.nz)
 * @version $Revision: 1.7 $
 */

public class VisualizeUtils {

  /** The name of the properties file */
  protected static String PROPERTY_FILE = "weka/gui/visualize/Visualize.props";

  /** Contains the visualization properties */
  protected static Properties VISUALIZE_PROPERTIES;

  /** Default maximum precision for the display of numeric values */
  protected static int MAX_PRECISION = 10;

  static {

    
    try {
      VISUALIZE_PROPERTIES = Utils.readProperties(PROPERTY_FILE);
      String precision = 
	VISUALIZE_PROPERTIES.getProperty("weka.gui.visualize.precision");
      if (precision == null) {
	/*
	System.err.println(weka.LocalString.get("Warning: no configuration property found in")
			   +PROPERTY_FILE
			   +weka.LocalString.get(" for weka.gui.visualize.precision. Using")
			   +weka.LocalString.get(" default instead."));*/
      } else {
	MAX_PRECISION = Integer.parseInt(precision);
	// System.err.println(weka.LocalString.get("Setting numeric precision to: ")+precision);
      }
    } catch (Exception ex) {
      JOptionPane.showMessageDialog(null,
       weka.LocalString.get("VisualizeUtils: Could not read a visualization configuration file.\n")
       +weka.LocalString.get("An example file is included in the Weka distribution.\n")
       +weka.LocalString.get("This file should be named \"") + PROPERTY_FILE + weka.LocalString.get("\"  and\n")
       +weka.LocalString.get("should be placed either in your user home (which is set\n")
       +"to \"" + System.getProperties().getProperty("user.home") + "\")\n"
       +weka.LocalString.get("or the directory that java was started from\n"),
       weka.LocalString.get("Plot2D"),
       JOptionPane.ERROR_MESSAGE);
    }
  }

  /**
   * Parses a string containing either a named colour or r,g,b values.
   * @param colourDef the string containing the named colour (or r,g,b)
   * @param defaultColour the colour to return if parsing fails
   * @return the Color corresponding to the string.
   */
  protected static Color processColour(String colourDef, Color defaultColour) {
    String colourDefBack = new String(colourDef);
    Color retC = defaultColour;
    if (colourDef.indexOf(",") >= 0) { 
      // Looks like property value is in R, G, B format
      try {
	int index = colourDef.indexOf(",");
	int R = Integer.parseInt(colourDef.substring(0,index));
	colourDef = colourDef.substring(index+1,colourDef.length());
	index = colourDef.indexOf(",");
	int G = Integer.parseInt(colourDef.substring(0,index));
	colourDef = colourDef.substring(index+1,colourDef.length());
	int B = Integer.parseInt(colourDef);
	//System.err.println(R+" "+G+" "+B);
	retC = new Color(R,G,B);
      } catch (Exception ex) {
	System.err.println(weka.LocalString.get("VisualizeUtils: Problem parsing colour property ")
			   +weka.LocalString.get("value (")+colourDefBack+").");
      }
    } else {
      // assume that the string is the name of a default Color.color
      if (colourDef.compareTo(weka.LocalString.get("black")) == 0) {
	retC = Color.black;
      } else if (colourDef.compareTo(weka.LocalString.get("blue")) == 0) {
	retC = Color.blue;
      } else if (colourDef.compareTo(weka.LocalString.get("cyan")) == 0) {
	retC = Color.cyan;
      } else if (colourDef.compareTo(weka.LocalString.get("darkGray")) == 0) {
	retC = Color.darkGray;
      } else if (colourDef.compareTo(weka.LocalString.get("gray")) == 0) {
	retC = Color.gray;
      } else if (colourDef.compareTo(weka.LocalString.get("green")) == 0) {
	retC = Color.green;
      } else if (colourDef.compareTo(weka.LocalString.get("lightGray")) == 0) {
	retC = Color.lightGray;
      } else if (colourDef.compareTo(weka.LocalString.get("magenta")) == 0) {
	retC = Color.magenta;
      } else if (colourDef.compareTo(weka.LocalString.get("orange")) == 0) {
	retC = Color.orange;
      } else if (colourDef.compareTo(weka.LocalString.get("pink")) == 0) {
	retC = Color.pink;
      } else if (colourDef.compareTo(weka.LocalString.get("red")) == 0) {
	retC = Color.red;
      } else if (colourDef.compareTo(weka.LocalString.get("white")) == 0) {
	retC = Color.white;
      } else if (colourDef.compareTo(weka.LocalString.get("yellow")) == 0) {
	retC = Color.yellow;
      } else {
	System.err.println(weka.LocalString.get("VisualizeUtils: colour property name not recognized ")
			   +"("+colourDefBack+").");
      }
    }
    return retC;
  }
}
